"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const experimental_utils_1 = require("@typescript-eslint/experimental-utils");
// eslint-disable-next-line @typescript-eslint/no-var-requires
const version = require('../package.json').version;
const createRule = experimental_utils_1.ESLintUtils.RuleCreator((name) => `https://github.com/0x706b/eslint-plugin-align-assignments/blob/v${version}/docs/rules/${name}.md`);
const spaceMatcher = /(\s*)((?!\['|"|`)(?!=>)(?:\+|-|\*|\/|%|&|&&|\^|\||\|\||<<|>>|\*\*|>>>)?=)(?!'|"|`\])/;
exports.default = createRule({
    name: 'align-assignments',
    meta: {
        type: 'layout',
        docs: {
            description: 'Enforce assignment alignment',
            recommended: false
        },
        fixable: 'code',
        messages: {
            notAligned: 'This group of assignments is not aligned'
        },
        schema: [
            {
                type: 'object',
                properties: {
                    maxColumns: {
                        type: 'number'
                    },
                    maxRows: {
                        type: 'number'
                    }
                }
            }
        ]
    },
    defaultOptions: [
        {
            maxRows: 1,
            maxColumns: 20
        }
    ],
    create(context, optionsWithDefaults) {
        const { maxRows, maxColumns } = optionsWithDefaults[0];
        const sourceCode = context.getSourceCode();
        const groups = [];
        let previousNode;
        return {
            VariableDeclaration(node) {
                var _a;
                if (previousNode &&
                    (previousNode === null || previousNode === void 0 ? void 0 : previousNode.type) === 'ExportNamedDeclaration' &&
                    ((_a = node.parent) === null || _a === void 0 ? void 0 : _a.type) === 'ExportNamedDeclaration') {
                    return;
                }
                addNode(node, node);
            },
            ExpressionStatement(node) {
                if (node.expression.type !== experimental_utils_1.AST_NODE_TYPES.AssignmentExpression) {
                    return;
                }
                addNode(node, node.expression);
            },
            ExportNamedDeclaration(node) {
                var _a;
                if (((_a = node.declaration) === null || _a === void 0 ? void 0 : _a.type) !== experimental_utils_1.AST_NODE_TYPES.VariableDeclaration) {
                    return;
                }
                addNode(node, node.declaration);
            },
            ClassDeclaration(node) {
                node.body.body.forEach((el) => {
                    if (el.type === experimental_utils_1.AST_NODE_TYPES.PropertyDefinition && el.value != null) {
                        addNode(el, el);
                    }
                });
            },
            'Program:exit': checkAll
        };
        function checkAll() {
            groups.forEach(check);
        }
        function isAssignmentExpression(node) {
            return node.type === experimental_utils_1.AST_NODE_TYPES.AssignmentExpression;
        }
        function isExportNamedDeclaration(node) {
            return node.type === experimental_utils_1.AST_NODE_TYPES.ExportNamedDeclaration;
        }
        function isExpressionStatement(node) {
            return node.type === experimental_utils_1.AST_NODE_TYPES.ExpressionStatement;
        }
        function isVariableDeclaration(node) {
            return node.type === experimental_utils_1.AST_NODE_TYPES.VariableDeclaration;
        }
        function addNode(groupNode, node) {
            if (shouldStartNewGroup(groupNode)) {
                groups.push([node]);
            }
            else {
                getLast(groups).push(node);
            }
            previousNode = groupNode;
        }
        function shouldStartNewGroup(node) {
            var _a, _b, _c;
            if (!previousNode) {
                return true;
            }
            if (node.parent !== previousNode.parent) {
                return true;
            }
            if (((_a = previousNode.parent) === null || _a === void 0 ? void 0 : _a.type) === 'ForStatement' &&
                previousNode.type === experimental_utils_1.AST_NODE_TYPES.VariableDeclaration &&
                previousNode.declarations) {
                return true;
            }
            const lastAssignmentNode = getLast(getLast(groups));
            if (lastAssignmentNode != null) {
                if (node.loc.start.line - lastAssignmentNode.loc.start.line > maxRows) {
                    return true;
                }
                if (Math.abs(findAssignment(lastAssignmentNode) - findAssignment(node)) > maxColumns) {
                    return true;
                }
            }
            const lineOfNode = (_b = sourceCode.getFirstToken(node)) === null || _b === void 0 ? void 0 : _b.loc.start.line;
            const lineOfPrev = (_c = sourceCode.getLastToken(previousNode)) === null || _c === void 0 ? void 0 : _c.loc.start.line;
            return lineOfNode != null && lineOfPrev != null ? lineOfNode - lineOfPrev !== 1 : true;
        }
        function check(group) {
            const maxPos = getMaxPos(group);
            if (maxPos && !areAligned(maxPos, group)) {
                context.report({
                    loc: {
                        start: group[0].loc.start,
                        end: getLast(group).loc.end
                    },
                    messageId: 'notAligned',
                    fix: (fixer) => {
                        const fixings = group.map((node) => {
                            const tokens = sourceCode.getTokens(node);
                            const firstToken = tokens[0];
                            const assignmentToken = tokens.find((token) => [
                                '=',
                                '+=',
                                '-=',
                                '*=',
                                '/=',
                                '%=',
                                '&=',
                                '^=',
                                '|=',
                                '>>=',
                                '<<=',
                                '**=',
                                '>>>=',
                                '||=',
                                '&&='
                            ].includes(token.value));
                            const line = sourceCode.getText(node);
                            const lineIsAligned = line.charAt(maxPos) === '=';
                            if (lineIsAligned || !assignmentToken || isMultiline(firstToken, assignmentToken)) {
                                return fixer.replaceText(node, line);
                            }
                            else {
                                const spacePrefix = firstToken.loc.start.column;
                                const startDelimiter = assignmentToken.loc.start.column - spacePrefix;
                                const endDelimiter = assignmentToken.loc.end.column - spacePrefix;
                                const start = line.slice(0, startDelimiter).replace(/\s+$/m, '');
                                const ending = line.slice(endDelimiter).replace(/^\s+/m, '');
                                const spacesRequired = maxPos - start.length - assignmentToken.value.length + 1;
                                const spaces = ' '.repeat(spacesRequired);
                                const fixedText = `${start}${spaces}${assignmentToken.value} ${ending}`;
                                return fixer.replaceText(node, fixedText);
                            }
                        });
                        return fixings.filter((fix) => fix);
                    }
                });
            }
        }
        function isMultiline(firstToken, assignmentToken) {
            return firstToken.loc.start.line !== assignmentToken.loc.start.line;
        }
        function findAssignment(node) {
            const prefix = getPrefix(node);
            const source = sourceCode.getText(node);
            const match = source.substr(prefix).match(spaceMatcher);
            const position = match && match.index != null ? match.index + prefix + match[2].length : null;
            return position;
        }
        function getPrefix(node) {
            let nodeBefore;
            if (isAssignmentExpression(node)) {
                nodeBefore = node.left;
            }
            else if (isVariableDeclaration(node)) {
                nodeBefore = node.declarations[0].id;
            }
            else if (isExpressionStatement(node)) {
                nodeBefore = node.expression.left;
            }
            else if (isExportNamedDeclaration(node)) {
                nodeBefore = node.declaration.declarations[0].id;
            }
            else {
                nodeBefore = node.key;
            }
            const prefix = nodeBefore.loc.end.column - nodeBefore.loc.start.column;
            return prefix;
        }
        function areAligned(maxPos, nodes) {
            return nodes
                .filter(assignmentOnFirstLine)
                .map((node) => sourceCode.getText(node))
                .every((source) => source.charAt(maxPos) === '=');
        }
        function getMaxPos(nodes) {
            const maxPos = nodes
                .filter(assignmentOnFirstLine)
                .map(findAssignment)
                .reduce((last, current) => Math.max(last, current), 0);
            return maxPos;
        }
        function assignmentOnFirstLine(node) {
            if (isAssignmentExpression(node)) {
                const onFirstLine = node.left.loc.start.line === node.right.loc.start.line;
                return onFirstLine;
            }
            else {
                const source = sourceCode.getText(node);
                const lines = source.split('\n');
                return lines[0].includes('=');
            }
        }
        function getLast(ary) {
            return ary[ary.length - 1];
        }
    }
});
